/*
 * Rjb - Ruby <-> Java Bridge
 * Copyright(c) 2004 Kuwashima
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * $Id: riconv.h 43 2007-12-26 18:55:04Z kuwa1 $
 */
#ifndef _RICONV_H
#define _RICONV_H


VALUE exticonv_local_to_utf8(VALUE);
VALUE exticonv_utf8_to_local(VALUE);

#endif /* _RICONV_H */
